package com.jdbernard.wdiwtlt.cli

import com.jdbernard.wdiwtlt.ConfigWrapper
import com.jdbernard.wdiwtlt.MediaLibrary
import com.jdbernard.wdiwtlt.db.ORM
import com.jdbernard.wdiwtlt.db.models.*
import com.jdbernard.io.NonBlockingInputStreamReader
import com.jdbernard.util.AnsiEscapeCodeSequence as ANSI
import com.zaxxer.hikari.HikariConfig
import com.zaxxer.hikari.HikariDataSource
import org.docopt.Docopt
import jline.console.ConsoleReader
import java.sql.Timestamp
import java.text.SimpleDateFormat
import java.net.URLDecoder
import java.util.regex.Pattern
import java.util.regex.Matcher
import org.slf4j.Logger
import org.slf4j.LoggerFactory
import uk.co.caprica.vlcj.discovery.NativeDiscovery
import uk.co.caprica.vlcj.component.AudioMediaListPlayerComponent
import uk.co.caprica.vlcj.player.MediaPlayerEventListener

import static com.jdbernard.util.AnsiEscapeCodeSequence.*

public class CommandLineInterface {

    public static final VERSION = "ALPHA"

    public static final def DOC = """\
wdiwtlt v$VERSION

Usage:
    wdiwtlt [options]
    wdiwtlt --version
    wdiwtlt --help

Options:
    -c --config <config-file>

        Config file for wdiwtlt CLI.

    -L --library-root <root-dir>

        The path to a local media library directory. Providing a local library
        causes wdiwtlt to run in local mode.

    -D --database-config <config-file>

        The path to a data source configuration file. wdiwtlt cli uses HikariCP
        for database connection pooling, so the given configuration file is
        expected to be formatted so that it can be fed directly to the
        HikariConfig constructor.
        (see https://github.com/brettwooldridge/HikariCP#configuration-knobs-baby)

Configuration:
"""

    private static Logger logger =
        LoggerFactory.getLogger(CommandLineInterface)

    private static UPPERCASE_PATTERN = Pattern.compile(/(.)(\p{javaUpperCase})/)

    private Properties cliConfig
    private MediaLibrary library

    /// IO Management
    private InputStream inStream
    private OutputStream outStream

    private ConsoleReader reader
    private Thread consoleReaderThread
    private List<String> consoleReadBuffer =
        Collections.synchronizedList(new ArrayList<String>())
    private synchronized boolean running

    /// Console output data
    private String titleStyle, normalStyle, statusStyle, promptStyle,
        artistStyle, albumStyle, fileStyle, errorStyle, playlistStyle

    private String clearLine = new ANSI().eraseLine(Erase.All).toString()
    private String afterInput =
        new ANSI().eraseLine(Erase.All).scrollUp().cursorUp().toString()
    private String beforeLeader =
        new ANSI().saveCursor().cursorPrevLine(3).toString()
    private String afterLeader =
        new ANSI().restoreCursor().eraseLine(Erase.ToEnd).toString()
    private String eraseLeader =
        new ANSI().eraseLine(Erase.All).cursorPrevLine().eraseLine(Erase.All)
            .cursorPrevLine().eraseLine(Erase.All)
            .cursorPrevLine().eraseLine(Erase.All).toString()

    public final static modelClass = [
        'album': Album, 'artist': Artist, 'bookmark': Bookmark,
        'mediaFile': MediaFile, 'playlist': Playlist, 'tag': Tag ]

    private int displayWidth = 79
    private long msgTimeout
    private ScrollText currentlyPlaying = new ScrollText(
        maxWidth: displayWidth - 17,
        text: "No media currently playing.")
    private ScrollText status = new ScrollText(maxWidth: displayWidth)
    private Date dismissMsgDate = new Date()
    private SimpleDateFormat sdf = new SimpleDateFormat('EEE-HH-SSS')

    /// Current play queue and selection data
    Selection currentSelection = new Selection()
    Playlist playQueue
    Bookmark playBookmark
    MediaFile curMediaFile

    /// VLCJ Player
    AudioMediaListPlayerComponent vlcj

    public static void main(String[] args) {

        def opts = new Docopt(DOC).withVersion("wdiwtlt v$VERSION").parse(args)

        def exitErr = { msg ->
            System.err.println("wdiwtlt: $msg")
            System.exit(1) }

        // Look for a given CLI config file.
        def givenCfg = new Properties()
        File cfgFile
        if (opts["--config"]) cfgFile = new File(opts["--config"])

        if (!cfgFile || !cfgFile.exists() || !cfgFile.isFile())
            cfgFile = new File("wdiwtlt.cli.properties")

        if (!cfgFile || !cfgFile.exists() || !cfgFile.isFile())
            cfgFile = new File(System.getenv().HOME, ".wdiwtlt.cli.properties")

        if (cfgFile.exists() && cfgFile.isFile()) {
            try { cfgFile.withInputStream { givenCfg.load(it) } }
            catch (Exception e) { givenCfg.clear() } }

        // Just in case, load up the general WDIWTLT default config service
        def wdiwtltDefaultConfig = new ConfigWrapper()

        // Find the library root directory (TODO: eventually this should be
        // enabled only when local mode is chosen).
        File libRoot = new File(
            opts["--library-root"] ?:
            givenCfg[ConfigWrapper.LIBRARY_DIR_KEY] ?:
            wdiwtltDefaultConfig.libraryRootPath ?:
            "no libRoot configured")

        if (libRoot && (!libRoot.exists() || !libRoot.isDirectory()))
            exitErr("Library root does not exist or is not a directory: " +
                libRoot.canonicalPath)

        // Get the library database
        HikariConfig hcfg
        File dbCfgFile = new File(
            opts["--database-config"] ?:
            givenCfg[ConfigWrapper.DB_CONFIG_FILE_KEY] ?:
            "database.properties")

        // Try to load from the given DB config file
        if (dbCfgFile && dbCfgFile.exists() && dbCfgFile.isFile()) {
            Properties props = new Properties()
            try { dbCfgFile.withInputStream { props.load(it) } }
            catch (Exception e) { props.clear() }

            if (props) hcfg = new HikariConfig(props) }

        // Look to see if database connection properties are given in the CLI
        // config file.
        if (givenCfg && givenCfg["database.config.dataSourceClassName"]) {
            Properties props = new Properties()
            props.putAll(givenCfg
                .findAll { it.key.startsWith("database.config.") }
                .collectEntries { [it.key[16..-1], it.value] } )
            if (props) hcfg = new HikariConfig(props) }

        // Fall back to the WDIWTLT general database defaults
        if (!hcfg) hcfg = wdiwtltDefaultConfig.hikariConfig
        if (!hcfg) exitErr("Cannot load database configuation.")

        // Create our database instance
        def orm
        try {
            def hds = new HikariDataSource(hcfg)
            orm = new ORM(hds) }
        catch (Exception e) {
            exitErr("Could not establish a connection to the database:\n\t" +
                e.localizedMessage) }

        // Try to discover the VLC native libraries
        def vlcj
        try {
            new NativeDiscovery().discover()
            vlcj = new AudioMediaListPlayerComponent() }
        catch (Exception e) {
            exitErr("Could not find the VLC libraries. Is VLC installed?\n\t" +
                e.localizedMessage) }

        // Create our CLI object
        try {
            def cliInst = new CommandLineInterface(new MediaLibrary(orm, libRoot),
                vlcj, System.in, System.out, givenCfg)
            cliInst.repl() }
        catch (Exception e) { e.printStackTrace(); exitErr(e.localizedMessage) }
        finally { if (vlcj) vlcj.release() }
    }

    public CommandLineInterface(MediaLibrary library,
    AudioMediaListPlayerComponent vlcj, InputStream sin, OutputStream out,
    Properties cliConfig) {
        this.cliConfig = cliConfig
        this.library = library
        this.inStream = sin
        this.outStream = out

        this.msgTimeout = cliConfig["message.timeout"] ?: 5000l

        setupTextStyles()

        library.clean()

        playQueue = library.save(new Playlist(
            name: "CLI Queue ${sdf.format(new Date())}"))

        String user = System.getProperty('user.name')
        String os = System.getProperty('os.name')
        def bookmarks = library.getBookmarkByName("Last played for $user on $os.")
        if (bookmarks) playBookmark = bookmarks[0]
        else playBookmark = new Bookmark(name: "Last played for $user on $os.")

        playBookmark.playlistId = playQueue.id
        playBookmark.playIndex = 0
        playBookmark.mediaFileId = 0

        this.vlcj = vlcj

        def listener = new UniversalNoopImplementation()
        listener.methods.playing = this.&playing
        listener.methods.finished = this.&finished
        listener.methods.stopped = this.&stopped
        vlcj.mediaListPlayer.mediaPlayer.addMediaPlayerEventListener(
            listener as MediaPlayerEventListener)

        this.consoleReaderThread = new Thread().start(this.&runReaderThread)

    }

    void setupTextStyles() {
        titleStyle = new ANSI().color(Colors.WHITE, Colors.DEFAULT, true).toString()
        normalStyle = new ANSI().resetText().toString()
        promptStyle = new ANSI().color(Colors.YELLOW, Colors.DEFAULT, true).toString()
        statusStyle = new ANSI().color(Colors.CYAN, Colors.DEFAULT, false).toString()
        playlistStyle = new ANSI().color(Colors.GREEN, Colors.DEFAULT, false).toString()
        artistStyle = new ANSI().color(Colors.RED, Colors.DEFAULT, false).toString()
        albumStyle = new ANSI().color(Colors.BLUE, Colors.DEFAULT, false).toString()
        fileStyle = new ANSI().color(Colors.GREEN, Colors.DEFAULT, false).toString()
        errorStyle = new ANSI().color(Colors.RED, Colors.DEFAULT, true).toString()
    }

    private void runReaderThread() {
        this.reader = new ConsoleReader(this.inStream, this.outStream)
        this.reader.setPrompt(promptStyle + "> " + normalStyle)

        String line
        while(this.running && !Thread.currentThread().isInterrupted()) {
            line = reader.readLine()
            if (line == null || line == "quit" || line == "exit") running = false
            else {
                consoleReadBuffer.add(line)
                outStream.print(afterInput)
                outStream.flush() } } }

    public void repl() {
        this.running = true
        String line = null

        outStream.println "\n\n\n"
        drawLeader()

        while(this.running) {

            if (new Date() > dismissMsgDate) resetStatus()
            if (consoleReadBuffer.size() > 0) {
                line = consoleReadBuffer.remove(0)
                line.split(';').each {
                    processInput(it.trim().split(/\s/) as LinkedList) }
            } else {
                drawLeader()
                if (curMediaFile) {
                    playBookmark.playTimeMs =
                        vlcj.mediaListPlayer.mediaPlayer.time
                    library.save(playBookmark) }
                Thread.sleep(250)
            }
        }
    }

    private def processInput(LinkedList<String> line) {
        logger.debug("line: $line")
        String command = line.poll()
        logger.debug("command: $command")
        switch(command?.toLowerCase()) {
            case 'scan':        return scanMediaLibrary()
            case 'select':      return processSelect(line, currentSelection)
            case 'list':        return processList(line, currentSelection)
            case 'add':         return processAdd(line)
            case 'enqueue':     return processEnqueue(line)
            case 'remove':      return processRemove(line)
            case 'tag':         return processTag(line)
            case 'clear':       return processClear(line)
            case 'play':        return processPlay(line)
            case 'pause':       return processPause(line)
            case 'stop':        return processStop(line)
            case 'next':        return processNext(line)
            case 'prev':        return processPrev(line)
            case 'jump':        return processJump(line)
            case 'ff':
            case 'fastforward': return processFastForward(line)
            case 'rw':
            case 'rwd':
            case 'rewind':      return processRewind(line)
            case 'vol':
            case 'volume':      return processVolume(line)

            case 'debug':
                outStream.println(
                    "\n\nConfig: \n" +
                    cliConfig.collect { "\t${it.key}: ${it.value}" }
                             .join("\n") +
                    "\n\n\n")
                drawLeader(true)
                return

            case 'q': case 'quit': case ':q': case 'exit': case '\u0004':
                running = false
                consoleReaderThread.interrupt()
                return

            default:
                status.text = errorStyle +
                    "Unrecognized command: '$line'${normalStyle}"
                dismissMsgDate = new Date(new Date().time + msgTimeout)
                drawLeader()
                Thread.sleep(250)
                break
        }
    }

    public MediaLibrary scanMediaLibrary() {
        status.text = "Scanning media library..."
        def counts = library.rescanLibrary()
        status.text = "Scanned ${counts.total} files. " +
            "Added ${counts.new} and ignored ${counts.ignored} files."
        dismissMsgDate = new Date(new Date().time + msgTimeout)
        return library }

    private def processSelect(LinkedList line, def sel = null) {
        String option = line.poll()
        boolean current = option == "current"
        if (!sel) sel = new Selection()
        def items

        if (current) {
            if (!curMediaFile) {
                setErr "No media is currently playing."
                return null }

            option = line.poll()
            switch (option) {
                case 'album':
                    sel.album = ensureExactlyOne(
                        library.getAlbumsWhere({ mediaFileId: curMediaFile.id}))
                    return sel
                case 'artist':
                    sel.artist = ensureExactlyOne(
                        library.getArtistsWhere({ mediaFileId: curMediaFile.id}))
                    return sel
                case 'playlist':
                    sel.playlist = playQueue; return sel
                case 'file': case 'mediaFile':
                    sel.mediaFile = curMediaFile; return sel
                case 'tags':
                    sel.tags = library.getTagsWhere({
                        mediaFileId: curMediaFile.id})
                    return sel
                default:
                    setErr("Unrecognized option to ${promptStyle}select " +
                        "current${errorStyle}.")
                    return null
            }
        }

        switch (option) {
            case 'file':
                sel.mediaFile = ensureExactlyOne(
                    library.getByIdOrName(MediaFile, line.join(' ')))
                return sel
            case 'playlist':
                sel.playlist = ensureExactlyOne(
                    library.getByIdOrName(Playlist, line.join(' ')))
                sel.playlist.lastUsed = new Timestamp(new Date().time)
                sel.playlist = library.save(sel.playlist)
                return sel
            case 'mediaFile': case 'album': case 'artist':
                sel[option] = ensureExactlyOne(
                    library.getByIdOrName(modelClass[option], line.join(' ')))
                return sel
            case 'tags':
                sel.tags = line.collect { library.getByIdOrName(Tag, it) }
                               .findAll().flatten()
                return sel

            default:
                setErr("Unrecognized option to ${promptStyle}select${errorStyle}")
                return null
        }
    }

    private def processAdd(LinkedList line) {

        List<String> parts = line.join(' ').split(' to playlist ').collect { it.trim() }

        if (parts.size() != 2) {
            printLongMessage(
                "Invalid options to the ${promptStyle}add${normalStyle}" +
                " command. Use ${promptStyle}help add${normalStyle} to see a " +
                "list of valid options.")
            return }

        Selection selectionToAdd
        Playlist targetPlaylist

        if (parts[0] == "selection") { selectionToAdd = currentSelection }
        else selectionToAdd = processSelect(parts[0].split(' ') as LinkedList)

        targetPlaylist = ensureExactlyOne(
            library.getByIdOrName(Playlist, parts[1]))

        return library.addToPlaylist(targetPlaylist.id,
            selection.selectedFiles.collect { it.id }) }

    private def processEnqueue(LinkedList line) {
        def selectedFiles = processSelect(line).selectedFiles
        selectedFiles.each {
            vlcj.mediaList.addMedia(
                new File(library.libraryRoot, it.filePath).canonicalPath) }
        playQueue = library.addToPlaylist(playQueue.id,
            selectedFiles.collect { it.id })
        return playQueue }

    private def processRemove(LinkedList line) {
        String[] parts = line.join(' ').split(' from ')
        Playlist playlist
        def selection

        if (parts.length == 1 || parts[1] == 'queue') playlist = playQueue
        else if (parts.length == 2) // TODO: shoule ensure starts with 'playlist'
            playlist = processSelect(parts[1].split(' ') as LinkedList)

        selection = processSelect(parts[0].split(' ') as LinkedList)

        if (selection) selection.selectedFiles.each { file ->
            library.removeFromPlaylist(playlist.id, file.id) } }

    private def processTag(LinkedList line) {
        List<String> parts = line.join(' ').split(' as ').collect { it.trim() }

        if (parts.size() != 2) {
            printLongMessage(
                "Invalid options to the ${promptStyle}tag${normalStyle}" +
                " command. Use ${promptStyle}help tag${normalStyle} to see a " +
                "list of valid options.")
            return }


        Selection selection = processSelect(parts[0].split(' ') as LinkedList)
        library.tagMediaFiles(selection.selectedFiles.collect { it.id },
            parts[1].split(' ') as List) }

    private def processClear(LinkedList line) {
        def option = line.poll()
        switch(option) {
            case null:
                print(new ANSI().eraseDisplay(Erase.All))//.cursorPosition(4, 0))
                drawLeader(true)
                return
            case 'queue':
                playQueue = library.removeAllFromPlaylist(playQueue.id)
                setPlayQueue(playQueue)
                return playQueue
            case 'selected playlist':
                if (!currentSelection.playlist) {
                    printLongMessage("No playlist currently selected.")
                    return null }
                return library.removeAllFromPlaylist(selected.playlist.id)
            case 'playlist':
                def criteria = line.poll()
                if (!criteria) {
                    printLongMessage("Missing playlist id or name. Use " +
                        "${promptStyle}help clear${normalStyle} to see a " +
                        "list of valid options.")
                    return null }
                def playlist = library.getByIdOrName(Playlist, criteria)
                if (!playlist) {
                    printLongMessage("No playlist matches '${criteria}'.")
                    return null }
                return library.removeAllFromPlaylist(playlist.id)
            case 'selection':
                currentSelection = new Selection()
                break
            default:
                printLongMessage("Unrecognized option to the ${promptStyle}" +
                    "clear${normalStyle} command. Use ${promptStyle}help clear" +
                    "${normalStyle} to see a list of valid options.")
                return null
        }
    }

    private def processList(LinkedList options, def selection) {
        logger.debug("Listing things. Options: $options")
        if (!selection) selection = new Selection()
        def option = options.poll()
        boolean all = option == 'all'
        boolean current = option == 'current'
        if (all || current) option = options.poll()

        logger.debug("Option: $option")

        def list

        switch(option) {
            case 'albums':
                if (all) list = library.getAlbums()
                else if (current) list = library.getAlbumsWhere(
                    playlistId: playQueue.id)
                else if (selection.album) list = [selection.album]
                else list = library.getAlbumsWhere(
                    playlistId: selection.playlist?.id,
                    artistId: selection.artist?.id,
                    mediaFileId: selection.mediaFile?.id)

                String albumMatch = options?.join(" ")?.trim()
                if (albumMatch) list = list.findAll { it.name =~ albumMatch }
                printLongMessage(makeList(selection, Album, list, all,
                    { "${it.id}: ${it}" }))
                break

            case 'artists':
                if (all) list = library.getArtists()
                else if (current) list = library.getArtistsWhere(
                    playlistId: playQueue.id)
                else if (selection.artist) list = [selection.artist]
                else list = library.getArtistsWhere(
                    playlistId: selection.playlist?.id,
                    albumId: selection.album?.id,
                    mediaFileId: selection.mediaFile?.id)

                String artistMatch = options?.join(" ")?.trim()
                if (artistMatch) list = list.findAll { it.name =~ artistMatch }
                printLongMessage(makeList(selection, Artist, list, all,
                    { "${it.id}: ${it}" }))
                break

            case 'files':
            case 'selection':
                if (all) list = library.getMediaFiles()
                else if (current) list = library.getMediaFilesWhere(
                    playlistId: playQueue.id)
                else list = selection.selectedFiles

                if (selection.album) list = list.sort { it.trackNumber }
                String mediaFileMatch = options?.join(" ")?.trim()
                if (mediaFileMatch) list = list.findAll {
                    it.name =~ mediaFileMatch }
                printLongMessage(makeList(selection, MediaFile, list, all,
                    { "${it.id}: ${it.trackNumber} - ${it}" }))
                break

            case 'bookmarks':
                if (all) list = library.getBookmarks()
                else if (current) list = library.getBookmarksWhere(
                    playlistId: playQueue.id)
                else list = library.getBookmarksWhere(
                    playlistId: selection?.playlist?.id)

                String bookmarkMatch = options?.join(" ")?.trim()
                if (bookmarkMatch)
                     list = list.findAll { it.name =~ bookmarkMatch }
                printLongMessage(makeList(selection, Bookmark, list, all,
                    { "${it.id}: ${it} ${it.userCreated ? '' : ' (auto)'}" }))
                break

            case 'playlists':
                if (all) list = library.getPlaylists()
                else if (current) list = [playQueue]
                else list = library.getPlaylistsWhere(
                    artistId: selection?.artist?.id,
                    albumId: selection?.album?.id,
                    mediaFileId: selection?.mediaFile?.id)

                String playlistMatch = options?.join(" ")?.trim()
                if (playlistMatch)
                     list = list.findAll { it.name =~ playlistMatch }
                printLongMessage(makeList(selection, Playlist, list, all,
                    { "${it.id}: ${it} ${it.userCreated ? '' : ' (auto)'}" }))
                break

            case 'tags':
                if (all) list = library.getTags()
                else if (current) list = library.getTagsWhere(
                    playlistId: playQueue.id)
                else list = library.getTagsWhere(
                    playlistId: selection?.playlist?.id,
                    artistId: selection?.artist?.id,
                    albumId: selection?.album?.id)

                String tagMatch = options?.join(" ")?.trim()
                if (tagMatch) list = list.findAll { it.name =~ tagMatch }
                printLongMessage(makeList(selection, Tag, list, all,
                    { "${it.id}: ${it}" }))
                break

            default:
                printLongMessage("Unrecognized option to the ${promptStyle}" +
                    "list${normalStyle} command. Use ${promptStyle}help list" +
                    "${normalStyle} to see a list of valid options.")
                return null
        }

        resetStatus()
        return list
    }

    private def processPlay(LinkedList line) {

        def option = line.poll()
        if(option) switch(option) {
            case 'playlist':
                Playlist p = ensureExactlyOne(
                    library.getByIdOrName(Playlist, line.join(' ')))

                if (!p) { setErr('No matching playlist found.'); return }
                setPlayQueue(p)
                break

            case 'bookmark':
                Bookmark b = ensureExactlyOne(
                    library.getByIdOrName(Bookmark, line.join(' ')))
                if (!b) { setErr('No matching bookmark found.'); return }

                Playlist p = library.getPlaylistById(b.playlistId)
                if (!p) {
                    setErr('The playlist no longer exists for this bookmark')
                    return }

                setPlayQueue(p)
                vlcj.mediaListPlayer.playItem(b.playIndex)

                if (b.playTimeMs > 0)
                    vlcj.mediaListPlayer.mediaPlayer.time = b.playTimeMs
                break

            case 'album': case 'artist': case 'file': case 'tags':
                if (vlcj.mediaListPlayer.isPlaying())
                    vlcj.mediaListPlayer.stop()

                vlcj.mediaList.clear()

                // If we're currently playing a user's playlist, don't change
                // it, just create a new system playlist.
                if (playQueue.userCreated) {
                    playQueue = library.save(new Playlist(
                        name: "CLI Queue ${sdf.format(new Date())}")) }
                else library.removeAllFromPlaylist(playQueue.id)

                line.addFirst(option)
                processEnqueue(line)

            default:
                printLongMessage(
                    "Invalid options to the ${promptStyle}play${normalStyle}" +
                    " command. Use ${promptStyle}help play${normalStyle} to see a " +
                    "list of valid options.") }

        vlcj.mediaListPlayer.play() }

    private def processPause(LinkedList line) { vlcj.mediaListPlayer.pause() }

    private def processStop(LinkedList line) {
        // TODO
        vlcj.mediaListPlayer.stop() }

    private def processNext(LinkedList line) {
        def count = line.poll()
        try { count = count ? count as int : 1 }
        catch (Exception e) { setErr("$count is not a valid number"); count = 1 }
        vlcj.mediaListPlayer.stop()
        if ((playBookmark.playIndex + count) < vlcj.mediaList.size())
            vlcj.mediaListPlayer.playItem(playBookmark.playIndex + count) }

    private def processPrev(LinkedList line) {
        def count = line.poll()
        try { count = count ? count as int : 1 }
        catch (Exception e) { setErr("$count is not a valid number"); count = 1 }
        vlcj.mediaListPlayer.stop()
        if ((playBookmark.playIndex - count) >= 0)
            vlcj.mediaListPlayer.playItem(playBookmark.playIndex - count) }

    private def processJump(LinkedList line) {
        String errMsg =
            "Invalid options to the ${promptStyle}jump${normalStyle}" +
            " command. Use ${promptStyle}help jump${normalStyle} to see a " +
            "list of valid options."

        String to = line.poll()
        if (!to || to.trim() != "to") { printLongMessage(errMsg); return }

        def criteria = line.poll()
        if (!criteria) { printLongMessage(errMsg); return }

        def target = getExactlyOne(MediaFile, criteria)
        if (!target) return

        int index = library.getMediaFilesWhere(playlistId: playQueue.id)
            .findIndexOf { it.id == target.id }
        if (index < 0) {
            setErr "'$target' is not in the current play queue."; return }

        printLongMessage("""
            target: $target
            index: $index""")

        vlcj.mediaListPlayer.mediaPlayer.stop()
        vlcj.mediaListPlayer.playItem(index) }

    private def processFastForward(LinkedList line) {
        def amount = line.poll()
        def unit = line.poll()?.trim()

        if (!amount) { amount = "10"; unit = "s" }
        if (!unit) unit = 's'

        try { amount = amount as int }
        catch (Exception e) { setErr "$amount must be an integer."; return }

        switch (unit) {
            case 'ms': case 'millis': case 'millisecond': case 'milliseconds':
                vlcj.mediaListPlayer.mediaPlayer.skip(amount)
                break
            case 's': case 'sec': case 'second': case 'seconds':
                vlcj.mediaListPlayer.mediaPlayer.skip(amount * 1000)
                break
            case 'm': case 'min': case 'minute': case 'minutes':
                vlcj.mediaListPlayer.mediaPlayer.skip(amount * 60000)
                break
            default: setErr "$unit must be one of 'milliseconds' " +
                "(or 'millis' or 'ms'), 'seconds' (or 'sec' or 's'), or " +
                "'minutes' (or 'min' or 'm')"
                return } }

    private def processRewind(LinkedList line) {
        def amount = line.poll()
        def unit = line.poll()?.trim()

        if (!amount) { amount = "10"; unit = "s" }
        if (!unit) unit = 's'

        try { amount = -(amount as int) }
        catch (Exception e) { setErr "$amount must be an integer."; return }

        switch (unit) {
            case 'ms': case 'millis': case 'millisecond': case 'milliseconds':
                vlcj.mediaListPlayer.mediaPlayer.skip(amount)
                break
            case 's': case 'sec': case 'second': case 'seconds':
                vlcj.mediaListPlayer.mediaPlayer.skip(amount * 1000)
                break
            case 'm': case 'min': case 'minute': case 'minutes':
                vlcj.mediaListPlayer.mediaPlayer.skip(amount * 60000)
                break
            default: setErr "$unit must be one of 'milliseconds' " +
                "(or 'millis' or 'ms'), 'seconds' (or 'sec' or 's'), or " +
                "'minutes' (or 'min' or 'm')"
                return } }

    private def processVolume(LinkedList line) {
        def percentage = line.poll()
        if (!percentage)
            setMsg("Volume: ${vlcj.mediaListPlayer.mediaPlayer.volume}")

        else {
            try { percentage = Math.min(Math.max(0, percentage as int), 200) }
            catch (Exception e) {
                setErr("Volume must be and integer between 0 and 200.")
                return }
            vlcj.mediaListPlayer.mediaPlayer.volume = percentage } }

    private void playing(def player) {
        try {
            def mediaFiles = library.getMediaFilesWhere(playlistId: playQueue.id)
            String absFilePath =
                URLDecoder.decode(vlcj.mediaListPlayer.currentMrl()[7..-1])

            def currentIdx = mediaFiles.findIndexOf {
                def mrlPath = new File(absFilePath).canonicalPath
                def mfPath = new File(library.libraryRoot, it.filePath).canonicalPath
                return mfPath == mrlPath }

            curMediaFile = mediaFiles[currentIdx]

            currentlyPlaying.text = makeFullMediaFileDescription(curMediaFile)

            playBookmark.playlistId = playQueue.id
            playBookmark.playIndex = currentIdx
            playBookmark.mediaFileId = curMediaFile.id

            library.save(playBookmark) }
        catch (Exception e) { printLongMessage(e.printStackTrace()) } }

    private void finished(def player) {
        curMediaFile = library.incrementPlayCount(curMediaFile)
        library.save(curMediaFile) }

    private void stopped(def player) {
        curMediaFile = null
        currentlyPlaying.text = "No media currently playing." }

    public void setPlayQueue(Playlist p) {
        if (vlcj.mediaListPlayer.isPlaying()) vlcj.mediaListPlayer.stop()
        playQueue = p

        vlcj.mediaList.clear()

        library.getMediaFilesWhere(playlistId: playQueue.id).each {
            vlcj.mediaList.addMedia(
                new File(library.libraryRoot, it.filePath).canonicalPath) } }

    public def ensureExactlyOne(def matches) {
        if (!matches) {
            setErr("Nothing matches.");
            return null }

        String englishName = toEnglish(matches[0].class.simpleName)
        if (matches.size() > 1) {
            setErr("Multiple  ${englishName}s match: " +
                matches.collect { "${it.id}: ${it.name}" }.join(', '))
            return null }

        return matches[0] }

    public def getExactlyOne(Class modelClass, def criteria) {
        return ensureExactlyOne(library.getByIdOrName(modelClass, criteria)) }

    private void drawLeader(afterOutput = false) {

        String leader = beforeLeader + getLeader() +
            (afterOutput ?
                (promptStyle + "> " + normalStyle) :
                afterLeader)

        outStream.print(leader)
        outStream.flush() }

    private void printLongMessage(String msg) {
        String result = new StringBuilder()
            .append(eraseLeader)
            .append(msg)
            .append("\n\n")
            .toString()

        outStream.println result
        drawLeader(true) }

    private String getLeader() {
        StringBuilder leader = new StringBuilder()
            .append(clearLine)
            .append(titleStyle)
            .append("WDIWTLT - v")
            .append(VERSION)
            .append('\n')

        StringBuilder statusLine = new StringBuilder()
            .append(statusStyle)

        if (playBookmark) statusLine.append('(')
            .append(playBookmark.playIndex + 1)
            .append('/')
            .append(vlcj.mediaList.size())
            .append(') ')

        statusLine.append(currentlyPlaying)

        if (curMediaFile) {
            long playTime = Math.floor(vlcj.mediaListPlayer.mediaPlayer.time / 1000) as long
            int statusLength = ANSI.strip(statusLine.toString()).length()
            statusLine.append((' ' * (displayWidth - (statusLength + 7))))
                .append(promptStyle)
                .append(String.format('%02d', Math.floor(playTime / 60) as int))
                .append(':')
                .append(String.format('%02d', Math.floor(playTime % 60) as int))
                .append(' ')}

        leader.append(clearLine)
            .append(statusLine)
            .append("\n")
            .append(clearLine)
            .append(normalStyle)
            .append(status)
            .append("\n")

        return leader.toString() }

    private String setMsg(String msg) {
        status.text = msg
        dismissMsgDate = new Date(new Date().time + msgTimeout) }

    private String setErr(String errMsg) {
        status.text = errorStyle + errMsg
        dismissMsgDate = new Date(new Date().time + msgTimeout) }

    private String makeFullMediaFileDescription(MediaFile mf) {
        def artist = library.getArtistsWhere( mediaFileId: mf.id)
        def album = library.getAlbumsWhere(mediaFileId: mf.id)

        StringBuilder s = new StringBuilder()

        if (artist) s.append(artistStyle)
            .append(artist[0])
            .append(normalStyle)
            .append(" / ")

        if (album) s.append(albumStyle)
            .append(album[0])
            .append(normalStyle)
            .append(" / ")

        s.append(fileStyle)
         .append(mf)
         .append(normalStyle)
         .append(' ')

        return s.toString() }

    private String makeList(Selection selection, Class modelClass,
    def items, boolean listAll = false, Closure toString = null) {

        if (!toString) toString = { it.toString() }

        def currentCollection

        if (curMediaFile) {
            def lookupFun

            switch (modelClass) {
                case Album: lookupFun = library.&getAlbumsWhere; break
                case Artist: lookupFun = library.&getArtistsWhere; break
                case Playlist: lookupFun = library.&getPlaylistsWhere; break
                case MediaFile: lookupFun = { ign -> [curMediaFile] }; break
                case Tag: lookupFun = library.&getTagsWhere; break
                default: lookupFun = { ign -> [] } }

            currentCollection = lookupFun([mediaFileId: curMediaFile.id]) }

        def highlightSelected = { item ->
            if (currentCollection && currentCollection.find {it.id == item.id})
                return "${promptStyle}${toString(item)}${normalStyle}"
            else return toString(item) }

        def result = new StringBuilder()
            .append("--------------------\n${modelClass.simpleName}s")

        if (!listAll && (selection.playlist || selection.artist ||
                         selection.mediaFile))
            result.append("\n(for selection: ")
                  .append(selection.toString())
                  .append(normalStyle)
                  .append(")")

        result.append(":\n\n")
        result.append(items.collect(highlightSelected).join("\n"))
            .append("\n\n")

        return result.toString() }

    private String resetStatus() {
        String s = currentSelection.toString()

        if (s.size() == 0) status.text = "No current media selections."
        else status.text = s

        return status.text }

    private static String uncapitalize(String s) {
        if (s == null) return null;
        if (s.length() < 2) return s.toLowerCase();
        return s[0].toLowerCase() + s[1..-1] }

    private static String toEnglish(String modelName) {
        return UPPERCASE_PATTERN.matcher(modelName).
            replaceAll(/$1 $2/).toLowerCase() }

    public class Selection {
        Album album
        Artist artist
        MediaFile mediaFile
        Playlist playlist
        List<Tag> tags

        public def unselect(Class modelClass) {
            String key = uncapitalize(modelClass.simpleName)
            def value = this[key]
            this[key] = null
            return value }

        public Selection select(Map s) {
            ['artist', 'album', 'playlist', 'file', 'tags'].each {
                this[it] = s[it] }
            return this
        }

        public def select(def value) {
            if (value) this[uncapitalize(match.class.simpleName)] = value
            return value }

        public List<MediaFile> getSelectedFiles() {
            if (album || artist || mediaFile || playlist || tags)
                return library.getMediaFilesWhere(
                    playlistId: playlist?.id,
                    artistId: artist?.id,
                    albumId: album?.id,
                    tags: tags)
            return null }

        public String toString() {
            StringBuilder s = new StringBuilder()

            if (playlist) s.append(playlistStyle)
                .append(playlist)
                .append(normalStyle)
                .append(": ")

            if (artist) s.append(artistStyle)
                .append(artist)
                .append(normalStyle)
                .append(" / ")

            if (album) s.append(albumStyle)
                .append(album)
                .append(normalStyle)
                .append(" / ")

            if (mediaFile) s.append(fileStyle)
                .append(mediaFile)
                .append(normalStyle)

            return s.toString()
        }


    }
}
