package com.jdbernard.wdiwtlt

import com.jdbernard.wdiwtlt.db.ORM
import com.jdbernard.wdiwtlt.db.models.*

import java.sql.Timestamp
import java.util.regex.Pattern

import org.apache.commons.codec.digest.DigestUtils
import org.jaudiotagger.audio.AudioFile
import org.jaudiotagger.audio.AudioFileIO
import org.jaudiotagger.tag.Tag as JATag
import org.jaudiotagger.tag.FieldKey
import org.slf4j.Logger
import org.slf4j.LoggerFactory

import static org.jaudiotagger.tag.FieldKey.*

public class MediaLibrary {

    private static Logger logger = LoggerFactory.getLogger(MediaLibrary)

    private static UPPERCASE_PATTERN = Pattern.compile(/(.)(\p{javaUpperCase})/)

    @Delegate ORM orm
    private File libraryRoot

    public long autoDeletePeriod = 1000 * 60 * 60 * 24 * 7 // one week

    public MediaLibrary(ORM orm, File rootDir) {
        logger.debug("Creating a MediaLibrary rooted at: {}",
            rootDir.canonicalPath)

        this.orm = orm
        this.libraryRoot = rootDir }

    public void clean() {
        orm.removeEmptyAlbums()
        orm.removeEmptyArtists()
        orm.removeEmptyPlaylists()

        Timestamp staleTS = new Timestamp(new Date().time - autoDeletePeriod)
        orm.withTransaction {
            def stalePlaylists = orm.getPlaylistsWhere(
                userCreated: false, lastUsedBefore: staleTS)
            stalePlaylists.each { orm.delete(it) } }
        orm.withTransaction {
            def staleBookmarks = orm.getBookmarksWhere(
                userCreated: false, lastUsedBefore: staleTS)
            staleBookmarks.each { orm.delete(it) } } }

    public def rescanLibrary() {
        def results = [ total: 0, ignored: 0, new: 0]

        Date startDate = new Date()
        libraryRoot.eachFileRecurse { file ->

            if (!file.isFile()) return
            def mf = addFile(file)

            results.total++
            if (!mf) results.ignored++
            else if (mf.dateAdded > startDate) results.new++ }

        return results }

    public MediaFile addFile(File f) {
        if (!f.exists() || !f.isFile()) {
            logger.info("Ignoring non-existant file: {}", f.canonicalPath)
            return null }

        def relPath = getRelativePath(libraryRoot, f)
        MediaFile mf = orm.getMediaFileByFilePath(relPath)

        if (mf) {
            logger.info(
                "Ignoring a media file I already know about: {}", relPath)
            return mf }

        // Read in the media's tags
        mf = new MediaFile()
        def af

        try { af = AudioFileIO.read(f) }
        catch (Exception e) {
            logger.info("Ignoring a file because I can't " +
                "read the media tag info:\n\t{}\n\t{}",
                f.canonicalPath, e.localizedMessage)
            return null }

        def fileTag = af.tag

        mf.name = fileTag?.getFirst(TITLE)?.trim() ?: f.name
        mf.filePath = relPath
        mf.comment = fileTag?.getAll(COMMENT)?.collect { it.trim() }?.join('\n\n')
        mf.discNumber = safeToInteger(fileTag?.getFirst(DISC_NO)) ?: 1
        mf.trackNumber = safeToInteger(fileTag?.getFirst(TRACK))

        def folderParts = mf.filePath.split("[\\\\/]")[1..<-1] as LinkedList

        // Find artist and album names (if any)
        def artistNames = fileTag?.getAll(ARTIST)?.collect { it.trim() }
        def albumNames = fileTag?.getAll(ALBUM)?.collect { it.trim() }

        if (!artistNames) {
            mf.metaInfoSource = MediaFile.FILE_LOCATION
            artistNames = folderParts.size() >= 2 ?  [folderParts[0]] : [] }

        if (!albumNames) {
            mf.metaInfoSource = MediaFile.FILE_LOCATION
            albumNames = [folderParts.peekLast()] }

        // Hash the file
        mf.fileHash = f.withInputStream { DigestUtils.md5Hex(it) }

        orm.withTransaction {
            orm.create(mf)
            associateWithArtistsAndAlbums(mf, artistNames, albumNames,
                safeToInteger(fileTag?.getFirst(YEAR))) }

        return mf
    }

    public def getByIdOrName(Class modelClass, String input) {
        def match
        if (safeToInteger(input)) {
            match = orm.getById(modelClass, safeToInteger(input))
            if (match) match = [match] }
        else {
            match = orm.getByName(modelClass, input)
            if (!match) match = orm.getLike(modelClass, ["name"], [input]) }
        return match }

    public List<Artist> getArtistsByName(String name) {
        return [orm.getArtistByName(name)] ?: orm.getArtistsLikeName(name) }

    public List<Album> getAlbumsByName(String name) {
        return orm.getAlbumsWhere(name: name) ?: orm.getAlbumsLikeName(name) }

    public List<MediaFile> collectMediaFiles(List<Model> models) {
        if (!models) return []

        return models.collectMany { m ->
            if (m.class == MediaFile) return [m]
            return orm.getMediaFilesWhere((idKeyFor(m.class)): m.id) }.findAll() }

    public List<Artist> splitArtist(Artist toSplit, Pattern splitPattern) {
        return splitArtist(toSplit, pattern.split(toSplit.name)) }

    public List<Artist> splitArtist(Artist toSplit, List<String> newNames) {
        def albums = orm.getAlbumsWhere(artistId: toSplit.id)
        def mediaFiles = orm.getMediaFilesByArtistId(toSplit.id)

        toSplit.name = newNames[0]
        List<Artist> newArtists = newNames[1..-1].collect { new Artist(it) }
        newArtists.each { newArtist ->
            albums.each { orm.associate(newArtist, it) }
            mediaFiles.each { orm.associate(newArtist, it) } }

        return [toSplit] + newArtists }

    private void associateWithArtistsAndAlbums(MediaFile mf,
    List<String> artistNames, List<String> albumNames, Integer albumYear) {

        // Find or create artists.
        def artists = artistNames.collect { artistName ->
            Artist artist = orm.getArtistByName(artistName)
            if (!artist) artist = orm.create(new Artist(name: artistName))
            return artist }

        // Associate file with artists.
        artists.each { orm.associate(it, mf) }

        // Find or create albums
        def albums = albumNames.collect { albumName ->
            Album album

            // If we know what year the album was released we can use that to
            // narrow down the list of matching albums.
            if (albumYear != null) {
                // Look first to see if we already know about this album
                // associated with one of the artists for this piece.
                album = artists.inject(null, { foundAlbum, artist ->
                    if (foundAlbum) return foundAlbum
                    def cur = orm.getAlbumsWhere(name: albumName,
                        year: albumYear, artistId: artist.id)
                    return cur ? cur[0] : null })

                // If we don't have it with one of the artists, see if we have
                // one that matches the name and year.
                if (!album) {
                    def cur = orm.getAlbumsWhere(
                        name: albumName, year: albumYear)
                    album = cur ? cur[0] : null } }

            else {
                album = artists.inject(null, { foundAlbum, artist ->
                    if (foundAlbum) return foundAlbum
                    def cur = orm.getAlbumsWhere(
                        name: albumName, artistId: artist.id)
                    return cur ? cur[0] : null })

                if (!album) {
                    def cur = orm.getAlbumsWhere(name: albumName)
                    album = cur ? cur[0] : null } }

            // We still can't find the album at all. We'll need to create it
            if (!album)
                album = orm.create(new Album(name: albumName, year: albumYear))

            return album }

        // Associate file with albums
        albums.each { orm.associate(it, mf) }

        // Make sure we have association between all of the artists and albums.
        artists.each { artist ->
            def albumsForArtist = orm.getAlbumsWhere(artistId: artist.id)
            def albumsMissing = albums - albumsForArtist
            albumsMissing.each { album -> orm.associate(artist, album) } } }

    /** #### `getRelativePath`
      * Given a parent path and a child path, assuming the child path is
      * contained within the parent path, return the relative path from the
      * parent to the child. */
    public static String getRelativePath(File parent, File child) {
        def parentPath = parent.canonicalPath.split("[\\\\/]")
        def childPath = child.canonicalPath.split("[\\\\/]")

        /// If the parent path is longer it cannot contain the child path and
        /// we cannot construct a relative path without backtracking.
        if (parentPath.length > childPath.length) return ""

        /// Compare the parent and child path up until the end of the parent
        /// path.
        int b = 0
        while (b < parentPath.length && parentPath[b] == childPath[b] ) b++

        /// If we stopped before reaching the end of the parent path it must be
        /// that the paths do not match. The parent cannot contain the child and
        /// we cannot build a relative path without backtracking.
        if (b != parentPath.length) return ""
        return (['.'] + childPath[b..<childPath.length]).join('/') }

    public static Integer safeToInteger(def val) {
        if (val == null) return null
        try { return val.trim() as Integer }
        catch (NumberFormatException nfe) { return null } }

    public static String uncapitalize(String s) {
        if (s == null) return null;
        if (s.length() < 2) return s.toLowerCase();
        return s[0].toLowerCase() + s[1..-1] }

    public static String toEnglish(Class c) {
        return UPPERCASE_PATTERN.matcher(c.simpleName).
            replaceAll(/$1 $2/).toLowerCase() }

    public static String idKeyFor(Class c) {
        return uncapitalize(c.simpleName) + 'Id'; }

}
